/**
 * EasyBeans
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: WarClassesLocator.java 4389 2008-12-15 13:48:57Z alitokmen $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.metadata.war.impl;

import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.archive.api.IArchiveManager;
import org.ow2.util.archive.impl.ArchiveManager;
import org.ow2.util.scan.api.IClassesLocator;
import org.ow2.util.scan.api.ScanException;
import org.ow2.util.scan.api.classlocator.ArchiveClassesLocator;

/**
 * This class is a temporary solution to solve war archive scan issue. Indeed an
 * instance of a class which implements ArchiveHelper should not be archive
 * dependent.
 * @author Gael Lalire
 */
public class WarClassesLocator implements IClassesLocator {

    /**
     * Class prefix.
     */
    private static final String CLASS_PREFIX = "WEB-INF/classes/";

    /**
     * Lib prefix.
     */
    private static final String LIB_PREFIX = "WEB-INF/lib/";

    /**
     * Lib extension.
     */
    private static final String LIB_SUFFIX = ".jar";

    /**
     * The war.
     */
    private IArchive archive;

    /**
     * List of library in WEB-INF/lib.
     */
    private List<IClassesLocator> libClassesLocator;

    /**
     * Constructor.
     * @param archive the war
     */
    public WarClassesLocator(final IArchive archive) {
        this.archive = archive;
    }

    /**
     * Create the library list.
     * @throws ArchiveException if an error occurs
     */
    private void init() throws ArchiveException {
        if (libClassesLocator != null) {
            return;
        }
        synchronized (this) {
            IArchiveManager archiveManager = ArchiveManager.getInstance();
            libClassesLocator = new ArrayList<IClassesLocator>();
            Iterator<String> iterator = archive.getEntries();
            while (iterator.hasNext()) {
                String entry = iterator.next();
                if (entry.startsWith(LIB_PREFIX) && entry.endsWith(LIB_SUFFIX)) {
                    libClassesLocator.add(new ArchiveClassesLocator(archiveManager.getArchive(archive.getResource(entry))));
                }
            }
        }
    }

    /**
     * @param className a class name
     * @return url of className inside the archive
     * @throws ScanException if error occurs
     */
    public URL getClassURL(final String className) throws ScanException {
        try {
            init();
        } catch (ArchiveException e) {
            throw new ScanException("A lib of a war cannot be analyzed", e);
        }
        String classLocation = className.replace('.', '/').concat(".class");
        URL url;
        try {
            url = archive.getResource(CLASS_PREFIX + classLocation);
        } catch (ArchiveException e) {
            throw new ScanException("The war cannot be analyzed", e);
        }
        if (url == null) {
            // search in WEB-INF/lib libs
            for (IClassesLocator classesLocator : libClassesLocator) {
                url = classesLocator.getClassURL(className);
                if (url != null) {
                    break;
                }
            }
        }
        return url;
    }

}
