/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: WarDeployableMetadataFactory.java 4645 2009-02-19 12:15:07Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.metadata.war.impl;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.ee.deploy.api.deployable.WARDeployable;
import org.ow2.util.ee.deploy.api.deployable.metadata.DeployableMetadataException;
import org.ow2.util.ee.metadata.common.impl.AbsDeployableMetadataFactory;
import org.ow2.util.ee.metadata.common.impl.xml.parsing.ParsingException;
import org.ow2.util.ee.metadata.war.api.IWarDeployableMetadata;
import org.ow2.util.ee.metadata.war.api.IWarDeployableMetadataFactory;
import org.ow2.util.ee.metadata.war.api.exceptions.WARDeploymentDescException;
import org.ow2.util.ee.metadata.war.impl.configurator.WarDeployableMetadataConfigurator;
import org.ow2.util.ee.metadata.war.impl.specific.ISpecificWarDeployableFactory;
import org.ow2.util.ee.metadata.war.impl.xml.merge.WarMetadataMerge;
import org.ow2.util.ee.metadata.war.impl.xml.struct.Filter;
import org.ow2.util.ee.metadata.war.impl.xml.struct.Listener;
import org.ow2.util.ee.metadata.war.impl.xml.struct.Servlet;
import org.ow2.util.ee.metadata.war.impl.xml.struct.Tag;
import org.ow2.util.ee.metadata.war.impl.xml.struct.WAR;
import org.ow2.util.ee.metadata.war.impl.xml.struct.WARDeploymentDesc;
import org.ow2.util.ee.metadata.ws.api.IWebservicesCommonClassMetadata;
import org.ow2.util.ee.metadata.ws.api.xml.merge.IClassMetadataFinder;
import org.ow2.util.ee.metadata.ws.api.xml.struct.IWebservices;
import org.ow2.util.ee.metadata.ws.impl.xml.merge.WebservicesMetadataMerger;
import org.ow2.util.ee.metadata.ws.impl.xml.parsing.Webservices12DeploymentDesc;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.scan.api.IClassesLocator;
import org.ow2.util.scan.api.IScanner;
import org.ow2.util.scan.api.classlocator.ClassLoaderClassesLocator;
import org.ow2.util.url.URLUtils;
import org.ow2.util.url.URLUtilsException;

/**
 * Factory to contruct war deployable metadata.
 * @author Gael Lalire
 */
public class WarDeployableMetadataFactory extends AbsDeployableMetadataFactory implements IWarDeployableMetadataFactory {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory.getLog(WarDeployableMetadataFactory.class);

    /**
     * Parsed classes, in a web-app are located in WEB-INF/classes.
     */
    private static final String WEBINF_CLASSES_PREFIX = "WEB-INF/classes/";

    /**
     * Default constructor.
     */
    public WarDeployableMetadataFactory() {
    }

    /**
     * Constructor with a scanner.
     * @param scanner the scanner which analyze class bytecode.
     */
    public WarDeployableMetadataFactory(final IScanner scanner) {
        super(scanner);
    }

    /**
     * Create war deployable metadata.
     * @param deployable the war deployable
     * @return the metadata
     * @throws DeployableMetadataException if error occurs
     */
    public IWarDeployableMetadata createDeployableMetadata(final WARDeployable deployable) throws DeployableMetadataException {
        return createDeployableMetadata(deployable, null, null);
    }

    /**
     * Create war deployable metadata.
     * @param deployable the war deployable
     * @param classLoader a classloader to use if a class is not in archive
     * @return the metadata
     * @throws DeployableMetadataException if error occurs
     */
    public IWarDeployableMetadata createDeployableMetadata(final WARDeployable deployable, final ClassLoader classLoader)
            throws DeployableMetadataException {
        List<? extends IClassesLocator> classesLocatorList = null;
        if (classLoader != null) {
            classesLocatorList = Arrays.asList(new ClassLoaderClassesLocator(classLoader));
        } else {
            classesLocatorList = Arrays.asList(new ClassLoaderClassesLocator(ClassLoader.getSystemClassLoader()));
        }
        return createDeployableMetadata(deployable, classesLocatorList, null);
    }

    /**
     * Extract a class list from xml.
     * @param war a object representation of web.xml
     * @param classNameSet a set to fill with class name inside web.xml
     */
    private void completeClassList(final WAR war, final Set<String> classNameSet) {

        // Add the servlets
        for (Servlet servlet : war.getServlets()) {
            addClassToSetIfNotNull(servlet.getClassName(), classNameSet);
        }

        // Add the listeners
        for (Listener listener : war.getListeners()) {
            addClassToSetIfNotNull(listener.getListenerClassName(), classNameSet);
        }

        // Add the tags
        for (Tag tag : war.getTags()) {
            addClassToSetIfNotNull(tag.getTagClassName(), classNameSet);
        }

        // Add the filters
        for (Filter filter : war.getFilters()) {
            addClassToSetIfNotNull(filter.getFilterClassName(), classNameSet);
        }

    }

    /**
     * Create war deployable metadata.
     * @param deployable the war deployable
     * @param warExternLibraries war extern libraries
     * @param specificDeployableFactoryList a specific deployable factory list
     * @return the metadata
     * @throws DeployableMetadataException if error occurs
     */
    public IWarDeployableMetadata createDeployableMetadata(final WARDeployable deployable,
            final List<? extends IClassesLocator> warExternLibraries,
            final List<? extends ISpecificWarDeployableFactory> specificDeployableFactoryList)
            throws DeployableMetadataException {
        try {
            if (!URLUtils.urlToFile2(deployable.getArchive().getURL()).isDirectory()) {
                throw new DeployableMetadataException("war must be uncompressed (local directory expected)");
            }
        } catch (URLUtilsException e) {
            throw new DeployableMetadataException("war must be uncompressed (local directory expected)");
        } catch (ArchiveException e) {
            throw new DeployableMetadataException("war exception", e);
        }

        // for time debugging
        long tAnalyzeStart = System.currentTimeMillis();

        IArchive archive = deployable.getArchive();

        // Get WAR DD.
        final WAR war;
        try {
            war = WARDeploymentDesc.getWAR(archive);
        } catch (WARDeploymentDescException e) {
            throw new DeployableMetadataException(e);
        }

        // Get Webservices DD.
        IWebservices webservices = null;
        try {
            webservices = Webservices12DeploymentDesc.getWebservices12(archive, Webservices12DeploymentDesc.Type.WEBAPP);
        } catch (ParsingException e) {
            throw new DeployableMetadataException(e);
        }

        WarDeployableMetadataConfigurator warArchiveMetadataConfigurator = new WarDeployableMetadataConfigurator(deployable);
        Set<String> classNameSet = new HashSet<String>();

        // call before scan on each specific
        if (specificDeployableFactoryList != null) {
            for (ISpecificWarDeployableFactory specificDeployableFactory : specificDeployableFactoryList) {
                warArchiveMetadataConfigurator
                        .addSpecificMetadataConfigurator(specificDeployableFactory.beforeScan(deployable, classNameSet));
            }
        }

        // First step : extract classes from XML
        completeClassList(war, classNameSet);

        // if metadata complete then ignore standard annotation (like @resource)
        if (war.isMetadataComplete()) {
            warArchiveMetadataConfigurator.setAnnotationParsingDesactived(true);
        } else {
            // enable annotation parsing because metadata descriptor is not
            // complete
            warArchiveMetadataConfigurator.setAnnotationParsingDesactived(false);

            Iterator<String> entryIterator;
            try {
                entryIterator = archive.getEntries();
            } catch (ArchiveException e) {
                throw new DeployableMetadataException(e);
            }
            while (entryIterator.hasNext()) {
                String entry = entryIterator.next();
                if (entry.startsWith(WEBINF_CLASSES_PREFIX) && entry.endsWith(CLASS_SUFFIX)) {
                    classNameSet.add(entry.substring(WEBINF_CLASSES_PREFIX.length(),
                                                     entry.length() - CLASS_SUFFIX.length()));
                }
            }
        }

        // do annotation scan
        scan(classNameSet, new WarClassesLocator(archive), warExternLibraries, warArchiveMetadataConfigurator);

        // Prepare merge
        final WarDeployableMetadata warDeployableMetadata = warArchiveMetadataConfigurator.getWarDeployableMetadata();

        // Merge with XML descriptor
        warDeployableMetadata.setWebservices12(webservices);
        new WarMetadataMerge(warDeployableMetadata).merge(war);
        WebservicesMetadataMerger.merge(warDeployableMetadata, new IClassMetadataFinder() {


            /**
             *  Servlet Link is a link to the servlet name
             */
            public IWebservicesCommonClassMetadata<?, ?, ?> findClassMetadata(final String link) {

                // No DD, do nothing
                if (war == null) {
                    return null;
                }

                // When the matching servlet is found, return the metadata of the servlet
                List<Servlet> servlets = war.getServlets();
                for (Servlet servlet : servlets) {
                    if (link.equals(servlet.getName())) {
                        return warDeployableMetadata.getWarClassMetadata(servlet.getClassName().replace('.', '/'));
                    }
                }

                // not found
                return null;
            }
        });

        // time if debugging
        if (LOGGER.isDebugEnabled()) {
            long tAnalyzeStartEnd = System.currentTimeMillis();
            LOGGER.debug("Analyze of file {0} took {1} ms.", archive.getName(), (tAnalyzeStartEnd - tAnalyzeStart));
        }

        return warArchiveMetadataConfigurator.getWarDeployableMetadata();
    }

    /**
     * @return the deployable metadata class
     */
    public Class<IWarDeployableMetadata> getDeployableMetadataClass() {
        return IWarDeployableMetadata.class;
    }

}
